/*
Copyright (C) 1996-2001  Ulric Eriksson <ulric@siag.nu>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL THE
AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/xpm.h>
#include <Mowitz/Mowitz.h>

static int find_file(const char *path, char *fn, const char *name)
{
	int found = 0;
	struct stat statbuf;
	char *p = NULL, *q;
	if (name[0] == '/') {
		strcpy(fn, name);
		found = !stat(fn, &statbuf);
		return found;
	} else {
		if (path) p = MwStrdup(path);
		else p = MwStrdup(DEFAULT_PIXPATH);
		for (q = strtok(p, ":"); q; q = strtok(NULL, ":")) {
			sprintf(fn, "%s/%s", q, name);
			found = !stat(fn, &statbuf);
			if (found) break;
		}
	}
	MwFree(p);
	return found;
}

Pixmap MwLoadPixmap(Display *dpy, Pixel color, const char *pm)
{
	XpmAttributes xa;
	Pixmap pm_return;
	int result;
	XpmColorSymbol symbol;
	char fn[1024];
	char *pixpath = getenv("PIXPATH");
	int found = 0;
	Screen *scr = DefaultScreenOfDisplay(dpy);

	found = find_file(pixpath, fn, pm);
	if (!found) {
		fprintf(stderr, "Pixmap %s not found in %s\n", pm, fn);
		return None;
	}

	xa.closeness = 40000;
	xa.exactColors = False;
	xa.valuemask = XpmCloseness | XpmExactColors;

	symbol.name = NULL;
	symbol.value = "none";
	symbol.pixel = color;
	xa.colorsymbols = &symbol;
	xa.numsymbols = 1;
	xa.valuemask |= XpmColorSymbols;

        result = XpmReadFileToPixmap(dpy,
                XRootWindowOfScreen(scr),
                fn, &pm_return, NULL, &xa);

        if (result != XpmSuccess) {
                fprintf(stderr, "XpmReadFileToPixmap(%s) returns %s\n",
                        fn, XpmGetErrorString(result));
                return None;
        }
        return pm_return;
}

/* color allocation stuff snarfed from the xmagick library */
#ifndef USHRT_MAX
#define USHRT_MAX 65535
#endif

static int rgb_shift = 10;

typedef struct xmagick_color {
        XColor color;
        unsigned long rgb;
} xmagick_color;

static xmagick_color known_colors[128];
static int ncolors, maxcolors = 128, lastcolor;

void MwFreeColors(void)
{    
        free(known_colors);
/*      known_colors = NULL;
*/
        ncolors = 0;
        lastcolor = -1;
}      

void MwAllocColor(Display *display, Colormap colormap, XColor *color)
{    
        int i, j, lower, upper;
        int cr, cg, cb, d;
        unsigned long rgb;
        cr = color->red >> rgb_shift;
        cg = color->green >> rgb_shift;
        cb = color->blue >> rgb_shift;
        rgb = (cr << 16) + (cg << 8) + cb;
       
        if (lastcolor != -1 && known_colors[lastcolor].rgb == rgb) {
                color->pixel = known_colors[lastcolor].color.pixel;
                return;
        }
        lower = 0;
        upper = ncolors-1;
        while (lower <= upper) {
                i = (lower+upper)/2;
                d = rgb-known_colors[i].rgb;
                if (d == 0) {
                        lastcolor = i;
                        color->pixel = known_colors[i].color.pixel;
                        return;
                } else if (d > 0) {
                        lower = i+1;
                } else {
                        upper = i-1;
                }
        }
        if (ncolors >= maxcolors) {
                color->pixel = 0;
                return;
        }
        color->flags = DoRed|DoGreen|DoBlue;
        if (XAllocColor(display, colormap, color) == 0) {
                int dr, dg, db, bd = INT_MAX, best = 0;
                for (j = 0; j < ncolors; j++) {
                        XColor c = known_colors[j].color;
                        dr = cr - (c.red >> rgb_shift);
                        dg = cg - (c.green >> rgb_shift);
                        db = cb - (c.blue >> rgb_shift);
                        d = dr*dr+dg*dg+db*db;
                        if (d < bd) {
                                bd = d;
                                best = j;
                        }
                }
                color->pixel = known_colors[best].color.pixel;
        }
        if (known_colors[upper].rgb > rgb) i = upper;
        else i = lower;
        if (i < 0) i = 0;
        if (i > ncolors) i = ncolors;
        for (j = ncolors; j > i; j--) {
                known_colors[j] = known_colors[j-1];
        }
        known_colors[i].color = *color;
        known_colors[i].rgb = rgb;
        known_colors[i].color.pixel = color->pixel;
        lastcolor = i;
        ncolors++;
}

void MwQueryColor(Display *display, Colormap colormap, XColor *color)
{
        int i;

        for (i = 0; i < ncolors; i++) {
                if (known_colors[i].color.pixel == color->pixel) {
                        color->red = known_colors[i].color.red;
                        color->green = known_colors[i].color.green;
                        color->blue = known_colors[i].color.blue;
                        return;
                }
        }
        if (i >= maxcolors) {
                color->red = 0;
                color->green = 0;
                color->blue = 0;
                return;
        }
        XQueryColor(display, colormap, color);
        known_colors[i].color = *color;
        ncolors++;
}


/* ---
Place the child centered relative to the parent.

011225: Center relative to screen instead.
*/

void MwCenter(Widget child)
{
        Position x, y;
        Dimension w1, h1;
	Window root;
	int rx, ry, rw, rh, rb, rd;

        XtRealizeWidget(child);
        XtUnrealizeWidget(child);
        XtVaGetValues(child,
                XtNwidth, &w1,
                XtNheight, &h1, (char *)0);
#if 0
        XtVaGetValues(parent,
                XtNwidth, &width,
                XtNheight, &height, (char *)0);
        XtTranslateCoords(parent,
                (Position)(width-w1)/2, (Position)(height-h1)/2, &x, &y);
#else
	XGetGeometry(XtDisplay(child), DefaultRootWindow(XtDisplay(child)),
		&root, &rx, &ry, &rw, &rh, &rb, &rd);
	x = (rw-w1)/2;
	y = (rh-h1)/2;
#endif
        XtVaSetValues(child,
                XtNx, x,
                XtNy, y, (char *)0);
}

String MwLabelGet(Widget w)
{
        String p = NULL;
        XtVaGetValues(w, XtNlabel, &p, (char *)0);
        return p;
}      

void MwLabelSet(Widget w, String p)
{
        char *q, *r;
        int must_set = 0;

        if (w == None || p == NULL) return;
        q = _((char *)p);
        if (q == NULL) return;
        r = MwLabelGet(w);

        if (r == NULL) {
                must_set = 1;
        } else {
                must_set = (strcmp(q, r) != 0);
        }

        if (must_set) XtVaSetValues(w, XtNlabel, q, (char *)0);
}

/* ---
Get the state of a toggle widget
*/

Boolean MwStateGet(Widget w)
{
        Boolean i = False;
        if (w != None)
                XtVaGetValues(w, XtNstate, &i, (char *)0);
        return i;
}

/* this is needed if we include vanilla Athena headers */
#ifndef XtNshadowWidth
#define XtNshadowWidth "shadowWidth"
#endif

/* ---
If no widget, do nothing.
If same state and same shadow width as before, do nothing.
If set, use shadow width ssw.
If unset, use shadow width usw.
*/

void MwStateSet(Widget w, Boolean i, int ssw, int usw)
{
        Boolean set;
        Dimension osw, nsw;
        if (w == None) return;
        XtVaGetValues(w,
                XtNstate, &set,
                XtNshadowWidth, &osw,
                (char *)0);
        if (i) nsw = ssw;
        else nsw = usw;
        if (i == set && nsw == osw) return;
        XtVaSetValues(w,
                XtNstate, i,
                XtNshadowWidth, nsw,
                (char *)0);
}

Dimension MwWidthGet(Widget w)
{
        Dimension x = 80;
        if (w != None)
                XtVaGetValues(w, XtNwidth, &x, (char *)0);
        return x;
}

/* ---
Get the height of a widget.
*/

Dimension MwHeightGet(Widget w)
{
        Dimension x = 20;
        if (w != None)
                XtVaGetValues(w, XtNheight, &x, (char *)0);
        return x;
}


int MwXErrorHandler(Display *dpy, XErrorEvent *event)
{
        char b[1000];

        XGetErrorText(dpy, event->error_code, b, 999);
        puts(b);
        return 0;
}

void MwSetIcon(Widget toplevel, char **icon_data)
{
	Pixmap icon, mask;
        XWMHints *hints;
        Display *dpy = XtDisplay(toplevel);
        int Width, Height;
        int scrn = DefaultScreen(dpy);
        int dpth = XDefaultDepth(dpy, scrn);
        Window win = DefaultRootWindow(dpy);
        /* Create the pixmaps accordingly                      *
         * All the pixmaps MUST have the same height and width */

        XpmAttributes xpmattributes;
        xpmattributes.depth = dpth;
        xpmattributes.valuemask = XpmSize | XpmDepth;
        XpmCreatePixmapFromData(dpy, win, icon_data, &icon, &mask,
                                &xpmattributes);
        Width = xpmattributes.width;
        Height = xpmattributes.height;
        
        /* Now tell the window manager what the icon is */
        hints = XAllocWMHints();
        hints->flags = IconPixmapHint | IconMaskHint;
        hints->icon_pixmap = icon;
        hints->icon_mask = mask;
        XSetWMHints(dpy, XtWindow(toplevel), hints);
        XFree(hints);
}

#define HIGHLIGHT_WIDTH 2
#define UNHIGHLIGHT_WIDTH 1

/* ---
This function highlights by increasing the shadow width. This
doesn't work for plain Athena widgets which have no shadow at
all, so they won't be highlighted.
*/

static void highlight_action(Widget w, XEvent *event,
        String *params, Cardinal *num_params)
{
        int sw;
        Boolean set;

        if (*num_params) sw = atoi(*params);
        else sw = HIGHLIGHT_WIDTH;
        if (*num_params > 1) {
                XtVaGetValues(w,
                        XtNstate, &set,
                        (char *)0);
        } else {
                set = 0;
        }
        if (set) {
                sw++;
        }

        if (w) {
                XtVaSetValues(w, "shadowWidth", sw, (char *)0);
        }
}

/* ---
Remove the highlighting.
*/

static void unhighlight_action(Widget w, XEvent *event,
        String *params, Cardinal *num_params)
{
        int sw;
        Boolean set;

        if (*num_params) sw = atoi(*params);
        else sw = UNHIGHLIGHT_WIDTH;
        if (*num_params > 1) {
                XtVaGetValues(w,
                        XtNstate, &set,
                        (char *)0);
        } else {
                set = 0;
        }
        if (set) {
                sw++;
        }

        if (w) {
                XtVaSetValues(w, "shadowWidth", sw, (char *)0);
        }
}

static XtActionsRec actions[] =
{
        {"siag-highlight", highlight_action},
        {"siag-unhighlight", unhighlight_action},
};

/* ---
Set the highlighting and unhighlighting colours.
*/

void MwHighlightInit(Widget pw)
{
        XtAppAddActions(XtWidgetToApplicationContext(pw),
                actions, XtNumber(actions));
}

